@@ -1,10 +1,10 @@
 from __future__ import annotations
+from geo3d.linalg import add_vec_vec, cast_vec_to_array, mult_mat_vec, norm_L2
 import numpy as np
 from scipy.spatial.transform import Rotation as R
-from math import sqrt
 from numba import njit
 from .auxiliary import html_table_from_matrix, html_table_from_vector
-from typing import Union, List, Tuple, Any, Sequence, TypeVar, Optional
+from typing import Union, Tuple, Sequence, Optional
 
 RotationMatrixLike = Union[Sequence[Sequence[float]], np.ndarray, "RotationMatrix"]
 VectorLike = Union[Sequence[float], np.ndarray, "Vector", "Point"]
@@ -514,7 +514,7 @@ class Point:
         if copy:
             obj._a = a.copy()  # storage as copied numpy array
         else:
-            obj._a = a # storage as copied numpy array passed by reference
+            obj._a = a  # storage as copied numpy array passed by reference
         return obj
 
     def express_in_frame(
@@ -894,54 +894,6 @@ def transform_points(
     # rotation can also be written as `np.einsum('ij,kj->ki', t0._rot, np.asarray(points))`
 
 
-@njit
-def normalized_quat(q) -> Tuple[float, float, float, float]:
-    """Return unit quaternion
-
-    by dividing by Euclidean (L2) norm
-
-    Args:
-        q array with elements
-            q0: quaternion component 0 (x)
-            q1: quaternion component 1 (y)
-            q2: quaternion component 2 (x)
-            q3: quaternion component 3 (scalar)
-
-    Returns:
-        array shape (4,) vector divided by L2 norm
-    """
-    n = norm_L2(q)
-    return (q[0] / n, q[1] / n, q[2] / n, q[3] / n)
-
-
-@njit
-def normalized_vector(vec) -> np.ndarray:
-    """Return unit vector
-
-    by dividing by Euclidean (L2) norm
-
-    Args:
-        vec array with elements x,y,z
-
-    Returns:
-        array shape (3,) vector divided by L2 norm
-    """
-    res = np.empty(3)
-    n = norm_L2(vec)
-    res[0] = vec[0] / n
-    res[1] = vec[1] / n
-    res[2] = vec[2] / n
-    return res
-
-
-@njit
-def norm_L2(vec) -> float:
-    s = 0
-    for v in vec:
-        s += v ** 2
-    return sqrt(s)
-
-
 @njit
 def quat_as_matrix(unit_quat):
     """Represent unit quaternion as rotation matrix.
@@ -999,29 +951,101 @@ def quat_as_matrix(unit_quat):
 
 
 @njit
-def add_vec_vec(v1, v2):
-    return (v1[0] + v2[0], v1[1] + v2[1], v1[2] + v2[2])
+def matrix_as_quat(matrix: np.ndarray):
+    """Rotation quaternion from rotation matrix.
+
+    Method from scipy.spatial.transform.Rotation,
+    jit-compiled by numba for speedup.
+
+    Rotations in 3 dimensions can be represented with 3 x 3 proper
+    orthogonal matrices [1]_. If the input is not proper orthogonal,
+    an approximation is created using the method described in [2]_.
+
+    Args:
+    matrix : array_like, shape (N, 3, 3) or (3, 3)
+        A single matrix or a stack of matrices, where ``matrix[i]`` is
+        the i-th matrix.
+
+    Returns:
+    quaternion array with elements
+            q0: quaternion component 0 (x)
+            q1: quaternion component 1 (y)
+            q2: quaternion component 2 (x)
+            q3: quaternion component 3 (scalar)
+
+    References
+    ----------
+    .. [1] https://en.wikipedia.org/wiki/Rotation_matrix#In_three_dimensions
+    .. [2] F. Landis Markley, "Unit Quaternion from Rotation Matrix",
+            Journal of guidance, control, and dynamics vol. 31.2, pp.
+            440-442, 2008.
+    """
+    # matrix = np.asarray(matrix, dtype=float)
+
+    decision_matrix = np.empty(4)
+    decision_matrix[0] = matrix[0][0]
+    decision_matrix[1] = matrix[1][1]
+    decision_matrix[2] = matrix[2][2]
+    decision_matrix[3] = decision_matrix[:3].sum()
+
+    choice = decision_matrix.argmax()
+
+    quat = np.empty(4)
+
+    if choice != 3:
+        i = choice
+        j = (i + 1) % 3
+        k = (j + 1) % 3
+
+        quat[i] = 1 - decision_matrix[3] + 2 * matrix[i, i]
+        quat[j] = matrix[j, i] + matrix[i, j]
+        quat[k] = matrix[k, i] + matrix[i, k]
+        quat[3] = matrix[k, j] - matrix[j, k]
+
+    else:
+        quat[0] = matrix[2, 1] - matrix[1, 2]
+        quat[1] = matrix[0, 2] - matrix[2, 0]
+        quat[2] = matrix[1, 0] - matrix[0, 1]
+        quat[3] = 1 + decision_matrix[3]
+
+    return normalized_quat(quat)
 
 
 @njit
-def mult_vec_sca(v, s):
-    return (v[0] * s, v[1] * s, v[2] * s)
+def normalized_vector(vec) -> np.ndarray:
+    """Return unit vector array
+
+    by dividing by Euclidean (L2) norm
+
+    Args:
+        vec array with elements x,y,z
+
+    Returns:
+        array shape (3,) vector divided by L2 norm
+    """
+    res = np.empty(3)
+    n = norm_L2(vec)
+    res[0] = vec[0] / n
+    res[1] = vec[1] / n
+    res[2] = vec[2] / n
+    return res
 
 
 @njit
-def dot_vec_vec(v1, v2):
-    return v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2]
+def normalized_quat(q) -> Tuple[float, float, float, float]:
+    """Return unit quaternion
 
+    by dividing by Euclidean (L2) norm
 
-@njit
-def mult_mat_vec(m, v):
-    return (dot_vec_vec(m[0], v), dot_vec_vec(m[1], v), dot_vec_vec(m[2], v))
+    Args:
+        q array with elements
+            q0: quaternion component 0 (x)
+            q1: quaternion component 1 (y)
+            q2: quaternion component 2 (x)
+            q3: quaternion component 3 (scalar)
 
-
-@njit
-def cast_vec_to_array(vec):
-    a = np.empty(3)
-    a[0] = vec[0]
-    a[1] = vec[1]
-    a[2] = vec[2]
-    return a
+    Returns:
+        array shape (4,) vector divided by L2 norm
+    """
+    n = norm_L2(q)
+    return (q[0] / n, q[1] / n, q[2] / n, q[3] / n)
